package ip1.ex14;

import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import java.net.URL;
import java.io.File;


/** JFtp is a multi-threaded graphical FTP client. The JFtp class implements
 *  the program's main thread which handles user itneraction and can fire up
 *  FTP control threads.
 *
 * @see JFtpThread
 * @see JFtpControl
 * 
 * @version $Id: JFtp.java,v 1.40 2003/10/17 14:03:39 sharky Exp $
 */
public class JFtp extends JFtpThread
{

  /** Program version. */
  protected final static String version = "0.8.4";
  /** JFtp's main window. */
  protected JFtpWindow mainWindow = null;

  /** Initial URL. */
  protected String argUrl = "ftp://dev.jerryweb.org/debian";
  /** Do we use passive mode ? */
  protected boolean argPassive = true;
  /** Do we use multithread? */
  protected int argMaxThreads = 1;

  /** The local working directory. */
  protected File localDir;


  /** The constructor parses the command line arguments.
   *
   * @param argv the command line arguments
   */
  public JFtp(String[] argv)
  {
    super(null,"JFtp",true);
    this.localDir = new File(".");

    // parse command line arguments
    int i = 0;
    while (i < argv.length) {
      String arg = argv[i++];

      if (arg.startsWith("-")) {
     // this is an option
        if (arg.equals("-h")) {
       // we want the program usage
          usage();
          System.exit(0);
        } else if (arg.equals("-p")) {
       // passive mode on
          argPassive = true;
        } else if (arg.equals("-a")) {
       // passive mode off
          argPassive = false;
        } else if (arg.equals("-s")) {
       // single control thread
          argMaxThreads = 1;
        } else if (arg.equals("-m")) {
       // mutliple control threads
          argMaxThreads = 5;
        } else {
       // invalid option
          debug("invalid option : " + arg);
          usage();
          System.exit(1);
        }
      } else {
     // this is not an option and we don't have an url
        argUrl = arg;
      }
    }

  }


  /** Displays a message both in applet and standard output.
   *
   * @param msg the message to print
   */
  public void debug(String msg)
  {
    System.out.println(msg);
    if (mainWindow != null)
      mainWindow.logArea.append(msg + '\n');
  }


  /** Download a file. */
  private boolean download(JFtpEntry entry)
  {
    JFtpControl auxControl = getAuxControl(entry.url);

    if (auxControl == null)
      return false;

    auxControl.queueRetr(entry);
    return true;
  }


  /** Delete a file.
   *
   * @param entry the entry to delete
   *
   * @see JFtpWindow#actionPerformed
   */
  protected void delete(JFtpEntry entry)
  {
    if (entry.type == JFtpEntry.DIR)
    {
       debug("cannot erase directory '" + entry.name + "'");
       return;
    }

    if (entry.location == JFtpEntry.LOCAL) {
      // if it's local, delete it on the spot
      File file = new File(localDir,entry.name);
      file.delete();
    } else {
      // for remote files, queue the request
      if (getControl() != null)
        getControl().queueDele(entry);
    }
  }


  /** Change the contents of the remote file list. */
  protected void dispRemote(Vector contents)
  {
    if (mainWindow != null)
    {
      mainWindow.remoteList.setListData(contents);
      mainWindow.repaint();
    }
  }


  /** Change the local working directory.
   *
   * @param dir the directory to change to
   */
  public void localCwd(String dir)
  {
    localDir = new File(localDir,dir);
    if (getControl() != null)
      getControl().localDir = localDir;
    refreshLocalList();
  }


  /** Return an auxilary control connection.
   *
   * @see #getControl
   */
  private JFtpControl getAuxControl(URL url)
  {
  // if we are above the max allowed threads, return null
    if (children.size() > argMaxThreads)
      return null;

  // try to locate an available thread
    JFtpControl auxControl = null;
    for (int i = 0; i < children.size(); i++) {
      if (!getChild(i).isBusy()) {
        auxControl = (JFtpControl) getChild(i);
        break;
      }
    }

  // if none is available, check whether we can fire one up
    if ( (auxControl == null)
      && (children.size() < argMaxThreads) )
    {
      auxControl = new JFtpControl(this,url);
      auxControl.start();
    }

    return auxControl;
  }


  /** Return the active control connection.
   *
   * @see #getAuxControl
   */
  protected JFtpControl getControl()
  {
    if (children.isEmpty())
      return null;

    return (JFtpControl) children.firstElement();
  }


  /** The program's main routine.
   *
   * @see #run
   */
  public static void main(String[] argv)
  {
    JFtp jftp = new JFtp(argv);
    jftp.run();
  }


  /** Connects to an FTP server.
   *
   * @param u the URL to connect to
   *
   * @see JFtpWindow#actionPerformed
   */
  protected void connect(URL u) {
    // drop whatever we where doing
    reset();

    JFtpControl newControl = new JFtpControl(this,u);
    if (mainWindow != null)
      mainWindow.logArea.setText("");
    newControl.start();
    newControl.queueList();
  }


  /** Queue the download of a file.
   *
   * @see #download
   * @see JFtpWindow#actionPerformed
   */
  public void queueDownload(JFtpEntry entry)
  {
    addQueue(new RootQueueItem(RootQueueItem.DOWNLOAD, entry));
  }


  /** Queue the upload of a file.
   *
   * @param entry the entry to upload
   *
   * @see #upload
   * @see JFtpWindow#actionPerformed
   */
  public void queueUpload(JFtpEntry entry)
  {
    addQueue(new RootQueueItem(RootQueueItem.UPLOAD, entry));
  }


  /** Remove a control thread.
   *
   * @see JFtpThread#removeChild
   * @see #refreshRemoteList
   */
  protected void removeChild(JFtpControl theThread)
  {
    // remove like any other JFtpThread
    super.removeChild(theThread);

    // if we have no children left
    if (getControl() == null)
      dispRemote(new Vector());
  }


  /** Refresh local file list.
   *
   * @see JFtpWindow#actionPerformed
   */
  protected void refreshLocalList()
  {
    if (mainWindow == null)
      return;

    mainWindow.localList.setListData(JFtpEntry.makeLocalList(localDir));
    mainWindow.repaint();
  }


  /** Refresh the remote file list.
   *
   * @see #dispRemote
   * @see JFtpControl#queueList
   */
  protected void refreshRemoteList()
  {
    if (getControl() == null)
      dispRemote(new Vector());
    else
      getControl().queueList();
  }


  /** Change the current URL in the window. */
  protected void refreshUrl()
  {
    if ( (mainWindow != null) && (getControl() != null) )
    {
      mainWindow.urlField.setText(getControl().remoteUrl.toString());
      mainWindow.repaint();
    }
  }


  /** Thread entry point. */
  public void run() {
    debug("JFtp ("+version+") firing up..");

    // fire up an instance of the JFtp class
    mainWindow = new JFtpWindow(this);
    refreshLocalList();

    // start a timer to update the thread view every 100ms
    Timer refreshTimer = new Timer();
    try { refreshTimer.schedule(new RefreshTask(this), 0, 100); }
    catch (IllegalStateException e) {}

    // main loop - process queue
    while (state != STATE_SHUTDOWN) {
      QueueItem item = peekQueue();
      if (item != null) {
        switch (item.action) {
          case QueueItem.SHUTDOWN:
            shutdown();
            break;
          case RootQueueItem.DOWNLOAD:
            if (download(item.getEntry()))
              readQueue();
            break;
          case RootQueueItem.UPLOAD:
            if (upload(item.getEntry()))
              readQueue();
            break;
        }
      }

      try {
        sleep(1);
      } catch (InterruptedException e){
      }
    }

    // stop the timer
    try { refreshTimer.cancel(); }
    catch (IllegalStateException e) {}

    debug("JFtp returning.");
    System.exit(0);
  }


  /** Set active or passive mode.
   */
  protected void setPassive(boolean passive)
  {
    argPassive = passive;
    if (getControl() != null)
      getControl().passive = argPassive;
  }


  /** Upload a file.
   *
   * @see #queueUpload
   */
  private boolean upload(JFtpEntry entry)
  {
    JFtpControl auxControl = getAuxControl(entry.url);

    if (auxControl == null)
      return false;

    auxControl.queueStor(entry);
    return true;
  }


  /** Display program usage. */
  private void usage() {
    debug("[ This is JFtp "+version+", a graphical ftp client ]\n\n"
           + "Syntax: \n"
           + "  JFtp [options] [url]\n\n"
           + " Options:\n"
           + "  -h\t- this help message\n"
           + "  -p\t- use passive mode [default]\n"
           + "  -a\t- use active mode\n"
           + "  -s\t- use single control thread [default]\n"
           + "  -m\t- use multiple control threads\n"
           );
  }


  /* Describes an item of JFtp's queue. */
  public class RootQueueItem extends QueueItem {
    public final static int DOWNLOAD       = 100;
    /** Upload a file. */
    public final static int UPLOAD         = 101;


    /** The constructor. */
    public RootQueueItem(int action, JFtpEntry entry) {
      super(action,entry);
    }


    /** Return string representation. */
    public String toString()
    {
      switch(action) {
      case DOWNLOAD:
        return "DOWNLOAD " + getEntry();
      case UPLOAD:
        return "UPLOAD " + getEntry();
      default:
        debug("RootQueueItem Unknown action.");
        return "";
      }
    }
  }


  /** A TimerTask to refresh JFtp's thread view.
   *
   * @see #refreshThreadView
   */
  class RefreshTask extends TimerTask {
    JFtp jftp;

    /** The constuctor.
     *
     * @param jftp the JFtp we send the refresh to
     */
    public RefreshTask(JFtp jftp) {
      this.jftp = jftp;
    }

    /** Triggered when the Timer reaches its limit.
     *
     * @see JFtp#run
     */
    public void run() {
      jftp.refreshThreadView();
    }
  }

}
