package ip1.ex14;

import java.io.*;
import java.net.MalformedURLException;
import java.net.Socket;
import java.net.URL;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;


/** The JFtpControl class implements an FTP control thread.
 *  Depending on the chosen threading model, at any moment there
 *  can be zero, one or any number of control threads running
 *  simultaneously. All control threads are owned by the JFtp
 *  instance. A control thread may own JFtpData threads.
 *
 *  @see JFtp
 *  @see JFtpData
 *  @see JFtpResponse
 *
 *  @version $Id: JFtpControl.java,v 1.42 2003/10/17 14:03:39 sharky Exp $
 */
public class JFtpControl extends JFtpThread
{
  /** Control connection. */
  protected Socket sockControl = null;
  /** We are connected .*/
  public final static int STATE_CONNECT = 1000;
  /** We have logged in. */
  public final static int STATE_LOGGED = 1100;

  /** The local working directory. */
  protected File localDir;
  /** The remote working URL. */
  protected URL remoteUrl;

  /** Are we in passive mode? */
  protected boolean passive;

  private BufferedReader in = null;
  private PrintWriter out = null;


  /** The constructor.
   *
   * @param jftp the JFtp paren thread
   * @param url the starting URL
   */
  public JFtpControl(JFtp jftp, URL url)
  {
    super(jftp,"JFtpControl");
    this.passive = jftp.argPassive;
    this.remoteUrl = url;
    this.localDir = jftp.localDir;
  }


  /** The command issued is not allowed in current state.
   *
   * @param cmd the rejected command
   */
  void die_command(String cmd) {
    debug("Command '"+cmd+"' not allowed in state '"+state+"'");
    shutdown();
  }


  /** The thread's entry point. */
  public void run()
  {
    parent.addChild(this);

    try {

      // connect, open reader and writer
      String host = remoteUrl.getHost();
      int atSign = host.indexOf('@');
      if (atSign != -1) host = host.substring(atSign+1);
      int port = remoteUrl.getPort();
      if (port == -1)
        port = remoteUrl.getDefaultPort();

      sockControl = new Socket(host,port);

      in = new BufferedReader(new InputStreamReader(sockControl.getInputStream()));
      out = new PrintWriter(sockControl.getOutputStream(),true);

      // get comments
      state = STATE_CONNECT;
      recvResponse();

      // send login and set start directory
      if (makeUser().isAck()) {
        state = STATE_LOGGED;
        String tmpPath = remoteUrl.getPath();
        if (!remoteUrl.getPath().equals(""))
          makeCwd(new JFtpEntry(remoteUrl.getPath(),JFtpEntry.REMOTE,JFtpEntry.DIR,null));
        else
          makePwd();
      } else {
        debug("Authentication failed, shutting down.");
        shutdown();
      }
    } catch (IOException e) {
      debug("Could not connect, shutting down.");
      shutdown();
    }

    Timer idleTimer = new Timer();
    int idleMain = 30 * 1000;
    int idleAux = 10 * 1000;

    // main loop - we process the queue
    while (state != STATE_SHUTDOWN) {
      QueueItem item = peekQueue();
      if (item != null) {
        // turn off the idle timer
        try { idleTimer.cancel(); }
        catch (IllegalStateException e) {}

        switch (item.action) {
          case ControlQueueItem.CWD:
            makeCwd(item.getEntry());
            break;
          case ControlQueueItem.DELE:
            makeDele(item.getEntry());
            break;
          case ControlQueueItem.LIST:
            makeList();
            break;
          case ControlQueueItem.NOOP:
            makeNoop();
            break;
          case ControlQueueItem.PWD:
            makePwd();
            break;
          case ControlQueueItem.RETR:
            makeRetr(item.getEntry());
            break;
          case ControlQueueItem.SHUTDOWN:
            shutdown();
            break;
          case ControlQueueItem.STOR:
            makeStor(item.getEntry());
            break;
        }
        // remove item from queue
        readQueue();

        // start the idle timer
        try {
          idleTimer = new Timer();
          if (( (JFtp)getRoot()).getControl() == this)
            idleTimer.schedule(new IdleTask(this), idleMain);
          else
            idleTimer.schedule(new IdleTask(this), idleAux);
        } catch (IllegalStateException e) {
        }

      }

      // sleep some..
      try { sleep(1); }
      catch (InterruptedException e){}
    }

    // stop the timer
    try { idleTimer.cancel(); }
    catch (IllegalStateException e) {}

    // exit main loop
    ((JFtp) parent).removeChild(this);
    state = STATE_EXITED;
  }


  /** Receive and process a reponse from the server.
   *
   * @see #sendCommand
   * @see JFtpResponse
   */
  protected JFtpResponse recvResponse()
  {
    try
    {
      synchronized(in) {
        return new JFtpResponse(this,in);
      }
    } catch (IOException e) {
      debug("::recvResponse caught IOException");
      return JFtpResponse.ioFailure();
    }
  }


  /** Sends a string to the remote host.
   *
   * @param strng the string to send
   *
   * @see #recvResponse
   */
  protected void sendCommand(String strng)
  {
    debug("---> " + strng);
    out.println(strng);
    out.flush();
  }


  /** Close reader, writer and socket.
   */
  protected void shutdown()
  {
    // clear queue and remove children
    reset();

    // log out
    makeQuit();

    if (out != null)
      out.close();
    if (in != null) {
      try { in.close(); }
      catch (IOException e) { }
    }
    if (sockControl != null) {
      try { sockControl.close(); }
      catch (IOException e) { }
    }

    state = STATE_SHUTDOWN;
  }


  /** Change the remote working directory.
   *
   * @param dir the directory to change to
   *
   * @see #queueCwd
   */
  private void makeCwd(JFtpEntry entry)
  {
    if (state != STATE_LOGGED)
      die_command("CWD");

    sendCommand("CWD " + entry.name);
    recvResponse();

    makePwd();
  }


  /** Delete a remote file.
   *
   * @param file the file to delete
   *
   * @see #queueDele
   */
  private void makeDele(JFtpEntry entry)
  {
    if (state != STATE_LOGGED)
      die_command("DELE");

    sendCommand("DELE " + entry.name);
    recvResponse();
  }


  /** Returns the directory listing.
   *
   * @see #queueList
   * @see JFtpData#queueReadStr
   * @see JFtp#dispRemote
   */
  private JFtpResponse makeList()
  {
    if (state != STATE_LOGGED)
      die_command("LIST");

    String lines[];
    JFtpResponse resp = JFtpResponse.ioFailure();


    try {
      // prepare a data thread
      StringBuffer sb = new StringBuffer();
      JFtpData sockData = new JFtpData(this);
      sockData.queueReadStr(sb);
      sockData.start();

      // send the list command and get the corresponding lines
      sendCommand("LIST -a");
      resp = recvResponse();
      if ( !resp.isAck() )
        debug(resp.toString());

      waitShutdown(sockData);
      lines = sb.toString().split("\n");
    } catch (IOException e) {
      debug("::makeList caught IOException");
      return resp;
    }

    Vector entries = JFtpEntry.makeRemoteList(remoteUrl,lines);
    ((JFtp)getRoot()).dispRemote(entries);

    // return the response
    return resp;
  }



  /** Send a NOOP command.
   *
   * @see #queueNoop
   * @see IdleTask
   */
  private JFtpResponse makeNoop()
  {
    if (state != STATE_LOGGED)
      die_command("NOOP");
    sendCommand("NOOP");
    return recvResponse();
  }


  /** Sent the user's password.
   */
  private JFtpResponse makePass()
  {
    if (state != STATE_CONNECT)
      die_command("PASS");

    //  password
    String pass = "jftp@";
    if (remoteUrl.getUserInfo() != null) {
      String userinfo = remoteUrl.getUserInfo();
      int colonSign = userinfo.indexOf(':');
      if (colonSign == -1) {
        pass = "";
      } else {
        pass = userinfo.substring(colonSign+1);
      }
    }

    sendCommand("PASS " + pass);

    return recvResponse();
  }


  /** Prints current working directory.
   *
   * @see #queuePwd
   */
  private JFtpResponse makePwd()
  {
    if (state != STATE_LOGGED)
      die_command("PWD");

    sendCommand("PWD");
    JFtpResponse resp = recvResponse();
    if ( !resp.isAck() ) {
      debug(resp.toString());
      return resp;
    }

    String cwd = resp.args.trim();
    int posSpace = cwd.indexOf(' ');
    if (posSpace != -1)
      cwd = cwd.substring(0,posSpace);
    // remove quotes
    if ((cwd.charAt(0) == '"') && (cwd.charAt(cwd.length()-1) == '"'))
      cwd = cwd.substring(1,cwd.length()-1);
    debug("---- Current working directory : " + cwd);
    try {
      String host = remoteUrl.toString();
      host = host.substring(0,host.length() - remoteUrl.getPath().length());
      remoteUrl = new URL(host + cwd);
      JFtp jftp = (JFtp) getRoot();
      jftp.refreshUrl();
    } catch (MalformedURLException e) {
      debug("::makePwd caught MalformedURLException");
    }
    return resp;
  }


  /** Retrieve a remote file.
   *
   * @param entry the file we are retrieving
   *
   * @see #makeType
   * @see #queueRetr
   * @see #recvResponse
   * @see #sendCommand
   * @see JFtpData#queueReceive
   */
  private void makeRetr(JFtpEntry entry)
  {
    if (state != STATE_LOGGED)
      die_command("RETR");

    try {
      // prepare a data thread
      JFtpData sockData = new JFtpData(this,entry.size);
      sockData.queueReceive(new File(localDir,entry.name));
      sockData.start();

      // set transfer type to binary
      if (!makeType("I")) {
        debug("Aborting transfer, could not set TYPE.");
        return;
      }

      // send the RETR command and get the corresponding lines
      sendCommand("RETR " + entry.name);
      JFtpResponse resp = recvResponse();

      if ( !resp.isAck() )
        debug(resp.toString());

      waitShutdown(sockData);
      ((JFtp) getRoot()).refreshLocalList();

    } catch (IOException e) {
      debug("::makeRetr caught IOException");
      return;
    }


  }


  /** Send a file to the server.
   *
   * @param file the local file we are sending
   *
   * @see #makeType
   * @see #queueStor
   * @see #recvResponse
   * @see #sendCommand
   * @see JFtpData#queueSend
   */
  private void makeStor(JFtpEntry entry)
  {
    if (state != STATE_LOGGED)
      die_command("STOR");

    try {
      // prepare a data thread
      File file = new File(localDir,entry.name);
      JFtpData sockData = new JFtpData(this,file.length());
      sockData.queueSend(file);
      sockData.start();

      // set transfer type to binary
      if (!makeType("I")) {
        debug("Aborting transfer, could not set TYPE.");
        return;
      }

      // send the STOR command and get the corresponding lines
      sendCommand("STOR "+entry.name);
      JFtpResponse resp = recvResponse();
      if ( !resp.isAck() )
        debug(resp.toString());

      waitShutdown(sockData);

      // queue a remote refresh
      ((JFtp) getRoot()).refreshRemoteList();
    } catch (IOException e) {
      debug("::makeStor caught IOException");
      return;
    }

  }


  /** Set the transfer type.
   *
   * @param type the transfer type
   *
   * @see #makeRetr
   * @see #makeStor
   * @see #sendCommand
   * @see #recvResponse
   */
  private boolean makeType(String type)
  {
    if (state != STATE_LOGGED)
      die_command("TYPE");

    sendCommand("TYPE " + type);
    JFtpResponse resp = recvResponse();
    return resp.isAck();
  }


  /** Says goodbye to the FTP server.
   *
   * @see #sendCommand
   * @see #recvResponse
   */
  private void makeQuit()
  {
    if (state == STATE_LOGGED) {
      sendCommand("QUIT");
      recvResponse();
      state = STATE_CONNECT;
    }
  }


  /** Authenticate.
   *
   * @see #makePass
   */
  private JFtpResponse makeUser()
  {
    if (state != STATE_CONNECT)
      die_command("USER");

    //  user name and password
    String user = "anonymous";
    if (remoteUrl.getUserInfo() != null) {
      String userinfo = remoteUrl.getUserInfo();
      int colonSign = userinfo.indexOf(':');
      if (colonSign == -1) {
        user = userinfo;
      } else {
        user = userinfo.substring(0,colonSign);
      }
    }

    sendCommand("USER " + user);

    // check the response
    JFtpResponse resp = recvResponse();
    if (resp.isAck()) {
      state = STATE_LOGGED;
    } else if (resp.isIncomplete()) {
      resp = makePass();
    }

    return resp;
  }


  /** Queue a CWD command.
   *
   * @param dir the remote directory to change to
   *
   * @see #makeCwd
   */
  public void queueCwd(JFtpEntry entry)
  {
    addQueue(new ControlQueueItem(ControlQueueItem.CWD,entry));
  }


  /** Queue a DELE command.
   *
   * @param dir the remote directory to change to
   *
   * @see #makeDele
   * @see JFtpWindow#actionPerformed
   */
  public void queueDele(JFtpEntry entry)
  {
    addQueue(new ControlQueueItem(ControlQueueItem.DELE,entry));
  }


  /** Queue a LIST command.
   *
   * @see #makeList
   */
  public void queueList()
  {
    addQueue(new ControlQueueItem(ControlQueueItem.LIST,null));
  }


  /** Queue a NOOP command.
   *
   * @see #makeNoop
   */
  public void queueNoop()
  {
    addQueue(new ControlQueueItem(ControlQueueItem.NOOP,null));
  }


  /** Queue a PWD command.
   *
   * @see #makePwd
   */
  public void queuePwd()
  {
    addQueue(new ControlQueueItem(ControlQueueItem.PWD,null));
  }


  /** Queue a RETR command.
   *
   * @param file the local file to write to
   *
   * @see #makeRetr
   */
  public void queueRetr(JFtpEntry entry)
  {
    addQueue(new ControlQueueItem(ControlQueueItem.RETR,entry));
  }


  /** Queue a STOR command.
   *
   * @param file the local file to read from
   *
   * @see #makeStor
   */
  public void queueStor(JFtpEntry entry)
  {
    addQueue(new ControlQueueItem(ControlQueueItem.STOR,entry));
  }


  /** Class for describing a JFtpControl's queue. */
  class ControlQueueItem extends QueueItem {
    /** Send a LIST command. */
    public final static int LIST = 200;
    /** Send a RETR command. */
    public final static int RETR = 201;
    /** Send a STOR command. */
    public final static int STOR = 202;
    /** Send a CWD command. */
    public final static int CWD  = 203;
    /** Send a PWD command. */
    public final static int PWD  = 204;
    /** Send a PWD command. */
    public final static int DELE = 205;
    /** Send a NOOP command. */
    public final static int NOOP = 206;

    /** The constructor. */
    public ControlQueueItem(int action, JFtpEntry entry) {
      super(action, entry);
    }

    /** Return string representation. */
    public String toString()
    {
      switch(action) {
      case CWD:
        return "CWD " + getEntry();
      case DELE:
        return "DELE " + getEntry();
      case LIST:
        return "LIST";
      case NOOP:
        return "NOOP";
      case PWD:
        return "PWD";
      case RETR:
        return "RETR " + getEntry();
      case SHUTDOWN:
        return "SHUTDOWN";
      case STOR:
        return "STOR " + getEntry();
      default:
        debug("ControlQueueItem : Unknown action");
        return "";
      }
    }
  }


  /** A TimerTask to watch inactivity on a JFtpControlThread.
   *
   * @see #queueNoop
   * @see JFtpThread#queueShutdown
   */
  class IdleTask extends TimerTask {
    JFtpControl thread;
    JFtp jftp;

    /** The constuctor.
     *
     * @param thread the JFtpControl thread we are watching.
     */
    public IdleTask(JFtpControl thread) {
      this.thread = thread;
      this.jftp = (JFtp) thread.getRoot();
    }

    /** Triggered where the Timer reaches its limit.
     *
     * @see JFtpControl#run
     */
    public void run() {

      if (jftp.getControl() == thread)
        thread.queueNoop();
      else
        thread.queueShutdown();
    }
  }
}

