package ip1.ex14;

import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.ServerSocket;
import java.net.URL;
import java.io.*;


/** The JFtpData class implements an FTP data thread. It supports
 *  both active and passive data connections. A data thread is
 *  owned by a JFtpControl thread.
 *
 *  @see JFtpControl
 *  @version $Id: JFtpData.java,v 1.24 2003/10/17 14:03:39 sharky Exp $
 */
public class JFtpData extends JFtpThread
{
  /** String buffer. */
  protected StringBuffer sb;
  /** The data socket */
  protected Socket sock = null;
  /** Listening socket (for active mode) */
  protected ServerSocket sockListen = null;
  /** Where to connect to (for passive mode) */
  protected InetSocketAddress sockAddr = null;

  /** The progress indicator. */
  protected long progress = 0;
  /** The target size. */
  protected long targetSize = 0;

  /** We are attempting to connect (initial state). */
  public final static int STATE_CONNECTING = 10;
  /** We have a connection. */
  public final static int STATE_CONNECTED = 20;

  /** Are we in passive mode? */
  boolean passive;

  /** Size of the buffer for file transfers. */
  final static int charBufferSize = 8192;
  /** Character buffer for file transfers. */
  char[] charBuffer = new char[charBufferSize];


  /** Variant of the constructor.
   */
  JFtpData(JFtpControl parent, long targetSize)
  throws IOException
  {
    this(parent);
    this.targetSize = targetSize;
  }

  /** The constructor.
   */
  JFtpData(JFtpControl parent)
  throws IOException
  {
    super(parent, "JFtpData");
    this.passive = parent.passive;
    this.state = STATE_CONNECTING;

    if (passive) {
      parent.sendCommand("PASV");
      JFtpResponse resp = parent.recvResponse();
      if ( !resp.isAck() )
        throw new IOException();
      sockAddr = resp.getAddr();
    } else {
      int port = parent.sockControl.getLocalPort();
      while (sockListen == null) {
        try {
          port++;
          sockListen = new ServerSocket(port);
        } catch (IOException e) {
          sockListen = null;
        }
      }

      String cmd = "PORT ";
      cmd += parent.sockControl.getLocalAddress().getHostAddress().replace('.',',');
      int n1 = (port / 256);
      int n2 = port % 256;
      cmd += "," + n1 + "," + n2;
      parent.sendCommand(cmd);

      JFtpResponse resp = parent.recvResponse();
      if ( !resp.isAck() )
        throw new IOException();
    }
  }


  /** Queue receiving a file from the server.
   *
   * @param file the file we output to.
   */
  public void queueReceive(File file)
  {
    addQueue(new DataQueueItem(DataQueueItem.RECEIVE, file));
  }


  /** Queue reading to a StringBuffer.
   *
   * @param sb the StringBuffer to output to.
   *
   * @see JFtpControl#makeList
   */
  public void queueReadStr(StringBuffer sb)
  {
    this.sb = sb;
    addQueue(new DataQueueItem(DataQueueItem.READSTR,null));
  }


  /** Queue sending a local file to the server. */
  public void queueSend(File file)
  {
    addQueue(new DataQueueItem(DataQueueItem.SEND, file));
  }


  /** Establish the data connection. */
  public void run()
  {
    parent.addChild(this);

    try {
      if (passive == true) {
        debug("connecting data socket to ("+sockAddr.getAddress().getHostAddress()+") port " + sockAddr.getPort());
        sock = new Socket();
        sock.connect(sockAddr);
      } else {
        debug("accepting data connection on port " + sockListen.getLocalPort());
        sock = sockListen.accept();
      }
      state = STATE_CONNECTED;

      // main loop - the single message pump
      DataQueueItem item = (DataQueueItem) peekQueue();
      if (item != null)
      {
        int c;
        int readChars;
        BufferedReader remoteIn;

        switch (item.action) {
        case DataQueueItem.RECEIVE:
          // read until there is either nothing left to read
          // or we we get an abort
          remoteIn = new BufferedReader(new InputStreamReader(sock.getInputStream()));
          FileWriter localOut = new FileWriter(item.getFile());
          while ( ((readChars = remoteIn.read(charBuffer,0,charBufferSize)) != -1)
               && ( state != STATE_SHUTDOWN) )
          {
            localOut.write(charBuffer,0,readChars);
            progress += readChars;
            refreshNode();
          }
          localOut.close();
          remoteIn.close();
          break;
        case DataQueueItem.READSTR:
          // read until there is either nothing left to read
          // or we we get an abort
          remoteIn = new BufferedReader(new InputStreamReader(sock.getInputStream()));
          while ( ((readChars = remoteIn.read(charBuffer,0,charBufferSize)) != -1)
               && ( state != STATE_SHUTDOWN) )
          {
            sb.append(charBuffer,0,readChars);
            progress += readChars;
            refreshNode();
          }
          remoteIn.close();
          break;
        case DataQueueItem.SEND:
          // write until there is either nothing left to read
          // or we we get an abort
          BufferedWriter remoteOut = new BufferedWriter(new OutputStreamWriter(sock.getOutputStream()));
          FileReader localIn = new FileReader(item.getFile());
          while ( ((readChars = localIn.read(charBuffer,0,charBufferSize)) != -1)
               && ( state != STATE_SHUTDOWN) )
          {
            remoteOut.write(charBuffer,0,readChars);
            progress += readChars;
            refreshNode();
          }
          localIn.close();
          remoteOut.close();
          break;
        }

        // if read response unless we are shutting down
        if (state != STATE_SHUTDOWN) {
          JFtpResponse resp = ((JFtpControl) parent).recvResponse();
          if ( !resp.isAck() ) {
            debug("queued item failed!");
            debug(resp.toString());
          }
        }

        readQueue();
      } // there is something in the queue


    } catch (IOException e) {
      debug("caught IOException");
    }


    parent.removeChild(this);
    state = STATE_EXITED;
  }


  /** String representation. */
  public String toString()
  {
    String me =  type + "[" + id + "] : " + JFtpEntry.formatSize(progress);
    if (targetSize > 0)
      me += " of " + JFtpEntry.formatSize(targetSize);
    return me;
  }


  /** Describes an item of a JFTpData's queue.
   */
  class DataQueueItem extends QueueItem {
    public final static int READSTR   = 300;
    public final static int RECEIVE   = 301;
    public final static int SEND      = 302;

    /** The constuctor. */
    public DataQueueItem(int action, File file) {
      super(action, file);
    }

    /** Return string representation. */
    public String toString()
    {
      switch(action) {
      case RECEIVE:
        return "RECEIVE "+getFile();
      case READSTR:
        return "READ ASCII";
      case SEND:
        return "SEND "+getFile();
      case SHUTDOWN:
        return "SHUTDOWN";
      default:
        debug("DataQueueItem : UNKNOWN ACTION");
        return "";
      }
    }
  }

}

