package ip1.ex14;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.net.URL;

/** The JFtpResponse class represents an FTP protocol response.
 *  As defined by RFC 959, FTP consist of one or more lines of
 *  text each starting with a 3 digit code.
 *
 *  @see JFtpControl
 *  @see <a href="http://www.faqs.org/rfcs/rfc959.html">RFC 959</a>
 *
 *  @version $Id: JFtpResponse.java,v 1.8 2003/10/16 14:17:59 sharky Exp $
 */
public class JFtpResponse
{
  int code;
  int type;
  String msg = "";
  String args = "";


  /** Private constructor for internal use. */
  private JFtpResponse(int type,String args) {
    this.type = type;
    this.code = type*100;
    this.args = args;
  }


  /** The constructor.
   *
   * @param control the JFtpControl we are operating in
   * @param in the BufferedRead to read from
   */
  public JFtpResponse(JFtpControl control, BufferedReader in)
  throws IOException
  {
    // read the first line of response
    boolean more = false;
    String ln = in.readLine();
    msg = ln;
    control.debug("<--- " + ln);

    type = Integer.parseInt(ln.substring(0,1));
    code = Integer.parseInt(ln.substring(0,3));
    args = ln.substring(4);
    more = (ln.charAt(3) == '-');

    while (more) {
      ln = in.readLine();

      if (ln.length() > 0) {
      // display the received response      
        msg += ln + '\n';
        control.debug("<--- " + ln);
        args += ln.substring(4);
        more = (ln.charAt(3) == '-');
      }
    }
  }

  /** Fake reply signaling an I/O error. */
  public static JFtpResponse ioFailure() {
    return new JFtpResponse(9,"Socket I/O error");
  }

  /** Is the response a positive acknowledgment ? */
  public boolean isAck()
  {
    return ((type==1) || (type == 2));
  }

  /** Did we get an 'incomplete information' response? */
  public boolean isIncomplete()
  {
    return (type == 3);
  }

  /** Display some info about the type of error. */
  public String toString()
  {
    String info = "FTP response : " + code + " " + args + "\n";
    switch (type)
    {
    case 0:
      info += "000 These replies are purely informative and constitute neither a\n"
            + "positive nor a negative acknowledgment.";
      break;
    case 1:
      info += "1xx Informative replies to status inquiries.  These constitute a\n"
            + "positive acknowledgment to the status command.";
      break;
    case 2:
      info += "2xx Positive acknowledgment of previous command or other success-\n"
            + "ful action.";
      break;
    case 3:
      info += "3xx Incomplete information.  Activity cannot proceed without\n"
            + "further specification and input.";
      break;
    case 4:
      info += "4xx Unsuccessful reply.  The request is correctly specified but\n"
            + "the server is unsuccessful in correctly fulfilling it.";
      break;
    case 5:
      info += "5xx Incorrect or illegal command.  The command or its parameters\n"
            + "were invalid or incomplete from a syntactic viewpoint, or the\n"
            + "command is inconsistent with a previous command.  The command\n"
            + "in question has been completely ignored.";
      break;
    }
    return info;
  }

  
  /** Extract the SocketAddress from the response.
   *
   * @see JFtpData#JFtpData
   */
  public InetSocketAddress getAddr()
  throws UnknownHostException
  {
    // check for validity
    int pos1 = args.indexOf('(');
    int pos2 = args.lastIndexOf(')');

    if ((pos1==-1) || (pos2==-1) || (pos2<=pos1))
      throw new UnknownHostException();

    String bits[] = args.substring(pos1+1,pos2).split(",");

    if (bits.length != 6)
      throw new UnknownHostException();

    // extract host and port
    byte addr[] = new byte[4];
    for (int i = 0; i < 4; i++)
      addr[i] = (new Integer(bits[i])).byteValue();
    int port = Integer.parseInt(bits[4]) * 256 + Integer.parseInt(bits[5]);

    return new InetSocketAddress(InetAddress.getByAddress(addr),port);
  }
}
