package ip1.ex14;

import javax.swing.JTree;
import javax.swing.tree.*;
import java.util.Vector;
import java.io.File;
import java.net.URL;

/** A thread that has a message queue and children. This serves
 *  as the base for all of JFtp's threads. It provides mechanisms
 *  to push and pop messages to and from its message queue. It also
 *  provides support for the thread hierarchy as a JFtpThread has
 *  a parent (null for the top thread) and can have children.
 *  
 *  There is also support for displaying the running threads by
 *  assigning the threadTree member of the toplevel thread. Threads
 *  signal a change in their queue or their children to the toplevel
 *  thread and the thread view is updated recursively.
 *
 *  @see JFtp
 *  @see JFtpControl
 *  @see JFtpData
 *
 *  @version $Id: JFtpThread.java,v 1.18 2003/10/17 14:03:39 sharky Exp $
 */
public class JFtpThread extends Thread
{
  /** The thread is not running. */
  public final static int STATE_NOTRUNNING = -1000;
  /** We are shutting down. */
  public final static int STATE_SHUTDOWN = 5000;
  /** The thread has exited. */
  public final static int STATE_EXITED = 9000;

  /** The message queue. */
  protected Vector queue = new Vector();
  /** The thread's children. */
  protected Vector children = new Vector();
  /** The thread's parent. Set to null for the toplevel thread.*/
  protected JFtpThread parent = null;

  /** An optional JTree for displaying the thread's queue and children. */
  protected JTree threadTree = null;
  /** Has the threadTree changed since the last refresh? */
  protected boolean threadTreeChanged = true;

  /** Type of thread. */
  protected String type = "JFtpThread";

  /** Instance counter. */
  protected static int counter = 0;
  /** This thread's id. */
  protected int id;

  /** The thread's state. */
  public int state = STATE_NOTRUNNING;

  /** The thread's node. */
  protected DefaultMutableTreeNode node;
  private DefaultMutableTreeNode idleNode = new DefaultMutableTreeNode("(idle)");

  /** The constructor.
   *
   * @param parent the parent thread, or null
   * @param type type of this thread
   * @param withThreadTree do we want a thread tree associated with this thread?
   */
  public JFtpThread(JFtpThread parent, String type, boolean withThreadTree)
  {
    this.parent = parent;
    this.type = type;
    this.id = counter++;
    this.node = new DefaultMutableTreeNode(this);
    if (withThreadTree)
      threadTree = new JTree(node);
    else
      threadTree = null;
    refreshNode();
  }


  /** Shorter version of the constructor (not tree).
   *
   * @param parent the parent thread, or null
   * @param type type of this thread
   */
  public JFtpThread(JFtpThread parent, String type)
  {
    this(parent,type,false);
  }


  /** Add a child thread.
   *
   * @param theThread the thread to add
   */
  protected synchronized void addChild(JFtpThread theThread)
  {
    // figure out where we'll insert the corresponding node
    int pos = 0;
    if (!children.isEmpty())
      pos = node.getIndex(getChild(children.size() -1).node) + 1;

    // add the child
    children.add(theThread);

    // add the corresponding node
    node.insert(theThread.node,pos);
    refreshNode();

    // signal child fireup
    theThread.debug("firing up..");
  }


  /** Queue a command.
   *
   * @param item the command to queue
   */
  protected synchronized QueueItem addQueue(QueueItem item)
  {
    // add the item to queue
    queue.add(item);

    // add the corresponding node
    node.add(item.node);
    refreshNode();

    return item;
  }


  /** Output a debugging message.
   *
   * @param msg the message to print
   */
  protected void debug(String msg)
  {
    String fullMsg = type + "[" + id + "] " + msg;

    if (parent != null) {
      parent.debug(fullMsg);
    } else {
      System.out.println(fullMsg);
    }
  }

  protected synchronized void dumpQueue()
  {
    for (int i = 0; i < queue.size(); i++)
     System.out.println(queue.elementAt(i).toString());
  }


  /** Return the requested child. */
  protected JFtpThread getChild(int index)
  {
    return (JFtpThread) children.elementAt(index);
  }


  /** Return the root thread. */
  protected JFtpThread getRoot()
  {
    if (parent == null)
      return this;
    else
      return parent.getRoot();
  }


  /** Is this thread busy (that is are there messages in its queue)? */
  protected boolean isBusy()
  {
    return !(queue.isEmpty() && children.isEmpty());
  }


  /** Peek at the next item in the queue. */
  protected QueueItem peekQueue()
  {
    if (queue.isEmpty())
      return null;
    return (QueueItem) queue.elementAt(0);
  }


  /** Propagate a notice that the tree view needs a refresh. */
  public void propagateTreeChange()
  {
    threadTreeChanged = true;
    if (parent != null)
      parent.propagateTreeChange();
  }


  /** Queue a shutdown of the thread. */
  protected void queueShutdown()
  {
    addQueue(new QueueItem(QueueItem.SHUTDOWN,null));
  }


  /** Read the next item in the queue. */
  protected synchronized QueueItem readQueue()
  {
    if (queue.isEmpty())
      return null;

    // removed item from queue
    QueueItem item;
    item = (QueueItem) queue.remove(0);

    // remove correponding node
    node.remove(item.node);
    refreshNode();

    return item;
  }


  /** Refresh the node representing this thread. */
  protected void refreshNode()
  {
    // check whether we have/need the idleNode as child
    if (isBusy()) {
      if (node.isNodeChild(idleNode))
        node.remove(idleNode);
    } else {
      if (!node.isNodeChild(idleNode))
        node.add(idleNode);
    }
    propagateTreeChange();
  }


  /** Update the thread view */
  protected void refreshThreadView()
  {
    if ( (threadTree != null)
      && (threadTreeChanged) )
    {
      synchronized(this) {
        threadTree.setModel(new DefaultTreeModel(node));
        for (int i = 0; i < threadTree.getRowCount(); i++)
          threadTree.expandRow(i);
        threadTreeChanged = false;
        threadTree.repaint();
      }
    } // we have a treeview and it is not up to date
  }



  /** Remove a child thread then updates the toplevel thread view.
   *
   * @param theThread the child to remove
   */
  protected synchronized void removeChild(JFtpThread theThread)
  {
    // signal child return
    theThread.debug("returning.");

    // remove the child
    children.remove(theThread);

    // remove the corresponding node
    node.remove(theThread.node);
    refreshNode();
  }



  /** Clear the thread's queue and kill all its children.
   */
  protected void reset()
  {
    // clear the queue
    while (readQueue() != null) {}

    // if we have any running children, kill them
    for (int i =0; i < children.size(); i++)
      getChild(i).shutdown();
  }


  /** The run function, must be overriden to do anything useful. */
  public void run()
  {
    // this is where you would insert the message pump
    // which runs until the state becomes STATE_SHUTDOWN

    // signal that the thread has exited
    state = STATE_EXITED;
  }


  /** Perform a shutdown. We kill all child processes then
   *  set the state to STATE_SHUTDOWN.
   */
  protected void shutdown()
  {
    // clear queue and kill children
    reset();

    // signal we have a shutdown
    state = STATE_SHUTDOWN;
  }


  /** Convert to String representation. */
  public String toString()
  {
    return type + "["+id+"] : " + queue.size() + " queued commands, "
           + children.size() + " child threads";
  }


   /** Wait for a thread to exit.
   *
   * @param theThread the thread to wait for
   */
  protected void waitShutdown(JFtpThread theThread) {
    while ( theThread.state != STATE_EXITED)
    {
      try { sleep(1); }
      catch (InterruptedException e) { }
    }
  }



  /** An item of the action queue. */
  class QueueItem
  {
    /** The item code to perform. */
    protected int action;
    /** Optional argument. */
    protected Object arg = new Object();
    /** Shutdown. */
    public final static int SHUTDOWN = 99;
    /** The queue item's representation as a node. */
    public DefaultMutableTreeNode node;

   /** The constructor. */
    public QueueItem(int action, Object arg)
    {
      this.action = action;
      this.arg = arg;
      this.node = new DefaultMutableTreeNode(this);
    }


    /** Return string representation. */
    public String toString()
    {
      switch (action) {
      case SHUTDOWN:
        return "SHUTDOWN";
      default:
        debug("JFtpThread : UNKNOWN ACTION");
        return "";
      }
    }


    /** Object accessor to arg. */
    public JFtpEntry getEntry()
    {
      return (JFtpEntry) arg;
    }


    /** File accessor to arg. */
    public File getFile()
    {
      return (File) arg;
    }

    /** String accessor to arg. */
    public String getString()
    {
      return (String) arg;
    }

    /** URL accessor to arg. */
    public URL getUrl()
    {
      return (URL) arg;
    }

    /** Vector accessor to arg. */
    public Vector getVector()
    {
      return (Vector) arg;
    }
  }

}
